#import python modules
import numpy as np
import time

#import compiled fortran library
import lib

#import atom writing library
import atomwrite

import sys
import os


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.
Cut = 2.5


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut, r0
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [lib.calcenergy(Pos + NormDir * x, M, L, Cut, r0) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = lib.calcenergy(Pos + NormDir * dx, M, L, Cut,r0)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [lib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut,r0)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [lib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut,r0)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = lib.calcenergy(Pos + NormDir * d3, M, L, Cut,r0)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = lib.calcenergy(Pos + NormDir * d3, M, L, Cut,r0)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global N, M, L, Cut, relaxpdb, r0
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = lib.calcenergyforces(Pos, M, L, Cut, r0, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        print(str(OldPE)+"\n")
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = lib.calcenergyforces(Pos, M, L, Cut, r0, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
        relaxpdb.write(Pos)
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), dtype=float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos

def InitPositionsCenterToX(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Offset = 0.1 * (np.random.rand(3) - 0.5)
        Pos[i,:] += Offset
        i += 1
    return Pos

def InitPositionsCenterToPlane(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Pert1 = 1.e-5*(np.random.rand() - 0.5)
        Pert2 = 1.e-5*(np.random.rand() - 0.5)
        Pos[i,1] += Pert1
        Pos[i,2] += Pert2
        i += 1
    return Pos

def InitPositionsCenterToLine(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Pert1 = 1.e-1*(np.random.rand() - 0.5)
        Pos[i,1] += Pert1
        i += 1
    return Pos

def InitPositionsLineToLine(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Offset = L* (np.random.rand() - 0.5)
        Pos[i,1] += Offset
        i += 1
    return Pos

def InitPositionsLineToPlane(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Offset = L* (np.random.rand() - 0.5)
        Pert = 1.e-5*(np.random.rand() - 0.5)
        Pos[i,1] += Offset
        Pos[i,0] += Pert
        i += 1
    return Pos

def InitPositionsPlaneToX(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Offset1 = L* (np.random.rand() - 0.5)
        Offset2 = L* (np.random.rand() - 0.5)
        Pert = 1.e-5*(np.random.rand() - 0.5)
        Pos[i,1] += Offset1
        Pos[i,0] += Offset2
        Pos[i,2] += Pert
        i += 1
    return Pos

def InitPositionsLineToX(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    i=0
    while i < N:
        Offset = L* (np.random.rand() - 0.5)
        Pert1 = 1.e-5*(np.random.rand() - 0.5)
        Pert2 = 1.e-5*(np.random.rand() - 0.5)
        Pos[i,1] += Offset
        Pos[i,0] += Pert1
        Pos[i,2] += Pert2
        i += 1
    return Pos

def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = lib.calcenergyforces(Pos, M, L, Cut, r0, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

###### TRIALS ######

def run():
    global N, M, L, dt, T, Cut, r0, relaxpdb

    # init params
    RescaleFreq = 100
    N = 50
    rho = 0.08
    L = (N/rho)**(1./3.)


    # (commented) optional: test energies/structures for many trials of different N

    Set = [50]#(np.linspace(10, 500 , 50)).astype(int)
    f = []
    s = []  
   # for i,r0 in enumerate(Set):
   #     fhand = open("f"+str(i)+".txt", "w")
   #     shand = atomwrite.pdbfile("restate"+str(i)+".txt", "w")
   #     f.append(fhand)
   #     s.append(shand)
    #runpdb = atomwrite.pdbfile("run.pdb", L, Compressed = True)

    relaxpdb = atomwrite.pdbfile("relax.pdb", L, Compressed = True)
    L = (N/rho)**(1./3.)
    writesnap = atomwrite.pdbfile("restate.pdb", L, Compressed = False)   

    M = 1 # single chain
    
    j = -1
    r0 = 1
    for N in Set:
        j += 1
        print("here"+str(N))
        L = (N/rho)**(1./3.)

        m = 1
        n = 1
        # S = (m,n), specify case
        if m == 0 and n == 1:
            Pos = InitPositionsCenterToLine(N,L)   # S = (0,1)
        elif m == 0 and n ==2:
            Pos = InitPositionsCenterToPlane(N,L) # S = (0,2)
        elif m == 1 and n ==1:
            Pos = InitPositionsLineToPlane(N,L) # S = (1,1)


        # extra cases
        # "___ToX" signifies going to a disordered 3d struct.
        #Pos = InitPositions(N,L)
        #Pos = InitPositionsCenterToX(N,L)
        #Pos = InitPositionsLineToLine(N,L) 
        #Pos = InitPositionsLineToX(N,L)
        #Pos = InitPositionsPlaneToX(N,L)
        
        # execute energy minimization
        PE, Pos = ConjugateGradient(Pos, dx=0.001, EFracTolLS=1.e-8, EFracTolCG=1.e-12)

        # write final pos 
        writesnap.write(Pos)
        #s[j].write(Pos)

        # extra: test system by running with velocity rescaling

        #dt = 0.01
        #T = 1.
        #MaxSteps = 2000
        #Vel = InitVelocities(N,T)
        #Accel = InitAccel(Pos)
        #i = 0
        #while i < MaxSteps:
        #    i += 1

        #    # step
        #    Pos, Vel, Accel, KE, PE = lib.vvintegrate(Pos, Vel, Accel, M, L, Cut, r0, dt)
        #    if i % RescaleFreq == 0:
        #        Vel = RescaleVelocities(Vel, T)
        #    if j == 0: 
        #        runpdb.write(Pos)

        #    # write
        #    print "%d %11.4f %11.4f %11.4f" % (i, PE+KE, PE, KE)
        #    f[j].write(str(i) + " " + str(PE) + "\n")
                

# execute
if __name__ == '__main__':
    run()
